
.. DO NOT EDIT.
.. THIS FILE WAS AUTOMATICALLY GENERATED BY SPHINX-GALLERY.
.. TO MAKE CHANGES, EDIT THE SOURCE PYTHON FILE:
.. "auto_recipes/plot_exporter_recipes_oe_lr.py"
.. LINE NUMBERS ARE GIVEN BELOW.

.. only:: html

    .. note::
        :class: sphx-glr-download-link-note

        :ref:`Go to the end <sphx_glr_download_auto_recipes_plot_exporter_recipes_oe_lr.py>`
        to download the full example code.

.. rst-class:: sphx-glr-example-title

.. _sphx_glr_auto_recipes_plot_exporter_recipes_oe_lr.py:


.. _l-plot-torch-linreg-101-oe:

====================================
Linear Regression and export to ONNX
====================================

:epkg:`scikit-learn` and :epkg:`torch` to train a linear regression.

data
====

.. GENERATED FROM PYTHON SOURCE LINES 13-30

.. code-block:: Python


    import numpy as np
    from sklearn.datasets import make_regression
    from sklearn.linear_model import LinearRegression, SGDRegressor
    from sklearn.metrics import mean_squared_error, r2_score
    from sklearn.model_selection import train_test_split
    import torch
    from onnxruntime import InferenceSession
    from experimental_experiment.helpers import pretty_onnx
    from onnx_array_api.plotting.graphviz_helper import plot_dot


    X, y = make_regression(1000, n_features=5, noise=10.0, n_informative=2)
    print(X.shape, y.shape)

    X_train, X_test, y_train, y_test = train_test_split(X, y)





.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    (1000, 5) (1000,)




.. GENERATED FROM PYTHON SOURCE LINES 31-37

scikit-learn: the simple regression
===================================

.. math::

      A^* = (X'X)^{-1}X'Y

.. GENERATED FROM PYTHON SOURCE LINES 37-44

.. code-block:: Python



    clr = LinearRegression()
    clr.fit(X_train, y_train)

    print(f"coefficients: {clr.coef_}, {clr.intercept_}")





.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    coefficients: [-1.52678363e-01  4.81531186e+01  6.42954043e+01 -1.90152247e-02
     -2.10422973e-01], -0.0024843283844364628




.. GENERATED FROM PYTHON SOURCE LINES 45-47

Evaluation
==========

.. GENERATED FROM PYTHON SOURCE LINES 47-53

.. code-block:: Python


    y_pred = clr.predict(X_test)
    l2 = mean_squared_error(y_test, y_pred)
    r2 = r2_score(y_test, y_pred)
    print(f"LinearRegression: l2={l2}, r2={r2}")





.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    LinearRegression: l2=92.55085646195683, r2=0.9871892691280045




.. GENERATED FROM PYTHON SOURCE LINES 54-58

scikit-learn: SGD algorithm
===================================

SGD = Stochastic Gradient Descent

.. GENERATED FROM PYTHON SOURCE LINES 58-64

.. code-block:: Python


    clr = SGDRegressor(max_iter=5, verbose=1)
    clr.fit(X_train, y_train)

    print(f"coefficients: {clr.coef_}, {clr.intercept_}")





.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    -- Epoch 1
    Norm: 69.10, NNZs: 5, Bias: 0.294272, T: 750, Avg. loss: 687.174653
    Total training time: 0.00 seconds.
    -- Epoch 2
    Norm: 77.45, NNZs: 5, Bias: 0.130717, T: 1500, Avg. loss: 79.308657
    Total training time: 0.00 seconds.
    -- Epoch 3
    Norm: 79.51, NNZs: 5, Bias: -0.001988, T: 2250, Avg. loss: 58.927697
    Total training time: 0.00 seconds.
    -- Epoch 4
    Norm: 80.03, NNZs: 5, Bias: -0.057954, T: 3000, Avg. loss: 57.496833
    Total training time: 0.00 seconds.
    -- Epoch 5
    Norm: 80.36, NNZs: 5, Bias: -0.117119, T: 3750, Avg. loss: 57.316691
    Total training time: 0.00 seconds.
    /home/xadupre/vv/this312/lib/python3.12/site-packages/sklearn/linear_model/_stochastic_gradient.py:1603: ConvergenceWarning: Maximum number of iteration reached before convergence. Consider increasing max_iter to improve the fit.
      warnings.warn(
    coefficients: [-1.75637160e-01  4.83375481e+01  6.41925528e+01  1.36168917e-02
     -2.77860867e-01], [-0.11711857]




.. GENERATED FROM PYTHON SOURCE LINES 65-66

Evaluation

.. GENERATED FROM PYTHON SOURCE LINES 66-73

.. code-block:: Python


    y_pred = clr.predict(X_test)
    sl2 = mean_squared_error(y_test, y_pred)
    sr2 = r2_score(y_test, y_pred)
    print(f"SGDRegressor: sl2={sl2}, sr2={sr2}")






.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    SGDRegressor: sl2=92.0893419873538, sr2=0.9872531511703074




.. GENERATED FROM PYTHON SOURCE LINES 74-76

Linrar Regression with pytorch
==============================

.. GENERATED FROM PYTHON SOURCE LINES 76-125

.. code-block:: Python



    class TorchLinearRegression(torch.nn.Module):
        def __init__(self, n_dims: int, n_targets: int):
            super().__init__()
            self.linear = torch.nn.Linear(n_dims, n_targets)

        def forward(self, x):
            return self.linear(x)


    def train_loop(dataloader, model, loss_fn, optimizer):
        total_loss = 0.0

        # Set the model to training mode - important for batch normalization and dropout layers
        # Unnecessary in this situation but added for best practices
        model.train()
        for X, y in dataloader:
            # Compute prediction and loss
            pred = model(X)
            loss = loss_fn(pred.ravel(), y)

            # Backpropagation
            loss.backward()
            optimizer.step()
            optimizer.zero_grad()

            # training loss
            total_loss += loss

        return total_loss


    model = TorchLinearRegression(X_train.shape[1], 1)
    optimizer = torch.optim.SGD(model.parameters(), lr=1e-3)
    loss_fn = torch.nn.MSELoss()

    device = "cpu"
    model = model.to(device)
    dataset = torch.utils.data.TensorDataset(
        torch.Tensor(X_train).to(device), torch.Tensor(y_train).to(device)
    )
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=1)


    for i in range(5):
        loss = train_loop(dataloader, model, loss_fn, optimizer)
        print(f"iteration {i}, loss={loss}")





.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    iteration 0, loss=1643633.375
    iteration 1, loss=158838.421875
    iteration 2, loss=89816.515625
    iteration 3, loss=86460.359375
    iteration 4, loss=86275.578125




.. GENERATED FROM PYTHON SOURCE LINES 126-127

Let's check the error

.. GENERATED FROM PYTHON SOURCE LINES 127-133

.. code-block:: Python


    y_pred = model(torch.Tensor(X_test)).detach().numpy()
    tl2 = mean_squared_error(y_test, y_pred)
    tr2 = r2_score(y_test, y_pred)
    print(f"TorchLinearRegression: tl2={tl2}, tr2={tr2}")





.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    TorchLinearRegression: tl2=92.1086611070357, tr2=0.9872504770508839




.. GENERATED FROM PYTHON SOURCE LINES 134-135

And the coefficients.

.. GENERATED FROM PYTHON SOURCE LINES 135-141

.. code-block:: Python


    print("coefficients:")
    for p in model.parameters():
        print(p)






.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    coefficients:
    Parameter containing:
    tensor([[-0.4810, 47.9366, 64.2478, -0.3368, -0.4574]], requires_grad=True)
    Parameter containing:
    tensor([0.0190], requires_grad=True)




.. GENERATED FROM PYTHON SOURCE LINES 142-146

Conversion to ONNX
==================

Let's convert it to ONNX.

.. GENERATED FROM PYTHON SOURCE LINES 146-150

.. code-block:: Python


    ep = torch.onnx.export(model, (torch.Tensor(X_test[:2]),), dynamo=True)
    onx = ep.model_proto





.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    [torch.onnx] Obtain model graph for `TorchLinearRegression([...]` with `torch.export.export(..., strict=False)`...
    [torch.onnx] Obtain model graph for `TorchLinearRegression([...]` with `torch.export.export(..., strict=False)`... ✅
    [torch.onnx] Run decomposition...
    [torch.onnx] Run decomposition... ✅
    [torch.onnx] Translate the graph into ONNX...
    [torch.onnx] Translate the graph into ONNX... ✅




.. GENERATED FROM PYTHON SOURCE LINES 151-152

Let's check it is work.

.. GENERATED FROM PYTHON SOURCE LINES 152-157

.. code-block:: Python


    sess = InferenceSession(onx.SerializeToString(), providers=["CPUExecutionProvider"])
    res = sess.run(None, {"x": X_test.astype(np.float32)[:2]})
    print(res)





.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    [array([[ 80.546555],
           [-33.39189 ]], dtype=float32)]




.. GENERATED FROM PYTHON SOURCE LINES 158-159

And the model.

.. GENERATED FROM PYTHON SOURCE LINES 159-163

.. code-block:: Python


    plot_dot(onx)





.. image-sg:: /auto_recipes/images/sphx_glr_plot_exporter_recipes_oe_lr_001.png
   :alt: plot exporter recipes oe lr
   :srcset: /auto_recipes/images/sphx_glr_plot_exporter_recipes_oe_lr_001.png
   :class: sphx-glr-single-img





.. GENERATED FROM PYTHON SOURCE LINES 164-169

Optimization
============

By default, the exported model is not optimized and leaves many local functions.
They can be inlined and the model optimized with method `optimize`.

.. GENERATED FROM PYTHON SOURCE LINES 169-176

.. code-block:: Python


    ep.optimize()
    onx = ep.model_proto

    plot_dot(onx)





.. image-sg:: /auto_recipes/images/sphx_glr_plot_exporter_recipes_oe_lr_002.png
   :alt: plot exporter recipes oe lr
   :srcset: /auto_recipes/images/sphx_glr_plot_exporter_recipes_oe_lr_002.png
   :class: sphx-glr-single-img





.. GENERATED FROM PYTHON SOURCE LINES 177-182

With dynamic shapes
===================

The dynamic shapes are used by :func:`torch.export.export` and must
follow the convention described there.

.. GENERATED FROM PYTHON SOURCE LINES 182-193

.. code-block:: Python


    ep = torch.onnx.export(
        model,
        (torch.Tensor(X_test[:2]),),
        dynamic_shapes={"x": {0: torch.export.Dim("batch")}},
        dynamo=True,
    )
    ep.optimize()
    onx = ep.model_proto

    print(pretty_onnx(onx))




.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    [torch.onnx] Obtain model graph for `TorchLinearRegression([...]` with `torch.export.export(..., strict=False)`...
    [torch.onnx] Obtain model graph for `TorchLinearRegression([...]` with `torch.export.export(..., strict=False)`... ✅
    [torch.onnx] Run decomposition...
    [torch.onnx] Run decomposition... ✅
    [torch.onnx] Translate the graph into ONNX...
    [torch.onnx] Translate the graph into ONNX... ✅
    opset: domain='pkg.onnxscript.torch_lib.common' version=1
    opset: domain='' version=18
    input: name='x' type=dtype('float32') shape=['s0', 5]
    init: name='linear.bias' type=float32 shape=(1,) -- array([0.01901066], dtype=float32)
    Constant(value=[[-0.48103...) -> t
      Gemm(x, t, linear.bias, beta=1.00, transB=0, alpha=1.00, transA=0) -> addmm
    output: name='addmm' type=dtype('float32') shape=['s0', 1]





.. rst-class:: sphx-glr-timing

   **Total running time of the script:** (0 minutes 3.257 seconds)


.. _sphx_glr_download_auto_recipes_plot_exporter_recipes_oe_lr.py:

.. only:: html

  .. container:: sphx-glr-footer sphx-glr-footer-example

    .. container:: sphx-glr-download sphx-glr-download-jupyter

      :download:`Download Jupyter notebook: plot_exporter_recipes_oe_lr.ipynb <plot_exporter_recipes_oe_lr.ipynb>`

    .. container:: sphx-glr-download sphx-glr-download-python

      :download:`Download Python source code: plot_exporter_recipes_oe_lr.py <plot_exporter_recipes_oe_lr.py>`

    .. container:: sphx-glr-download sphx-glr-download-zip

      :download:`Download zipped: plot_exporter_recipes_oe_lr.zip <plot_exporter_recipes_oe_lr.zip>`


.. include:: plot_exporter_recipes_oe_lr.recommendations


.. only:: html

 .. rst-class:: sphx-glr-signature

    `Gallery generated by Sphinx-Gallery <https://sphinx-gallery.github.io>`_
